(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: Lighting.cpp
//
// Desc: Example code showing how to use D3D lights.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
unit LightingUnit;

interface

uses
  Windows, Messages,
  Direct3D9, D3DX9,
  D3DApp, D3DFont, D3DUtil, DXUtil,
  Resource;

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------

type
  // Custom D3D vertex format used by the vertex buffer
  PMyVertex = ^TMyVertex;
  TMyVertex = packed record
    p: TD3DXVector3;       // vertex position
    n: TD3DXVector3;       // vertex normal
  // static const DWORD FVF;
  end;
  
  PMyVertexArray = ^TMyVertexArray;
  TMyVertexArray = array [0..0] of TMyVertex;

const
  MYVERTEX_FVF = D3DFVF_XYZ or D3DFVF_NORMAL;

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
    m_pWallMesh: ID3DXMesh;            // Tessellated plane to serve as the walls and floor
    m_pSphereMesh: ID3DXMesh;          // Representation of point light
    m_pConeMesh: ID3DXMesh;            // Representation of dir/spot light
    m_pFont: CD3DFont;                 // Font for drawing text
    m_pFontSmall: CD3DFont;            // Font for drawing text
    m_light: TD3DLight9;               // Description of the D3D light
    m_n: LongWord;                     // Number of vertices in the wall mesh along X
    m_m: LongWord;                     // Number of vertices in the wall mesh along Z
    m_nTriangles: LongWord;            // Number of triangles in the wall mesh
    m_bShowHelp: Boolean;
    m_bWireframe: Boolean;

  protected
    function BuildWallMesh: HRESULT; 

    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function FrameMove: HRESULT; override;
    function Render: HRESULT; override;
    function FinalCleanup: HRESULT; override;

  public
    constructor Create;

    function MsgProc(hWnd: HWND; uMsg: LongWord; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;



implementation

uses SysUtils;

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;
  m_strWindowTitle            := 'Lighting';
  m_d3dEnumeration.AppUsesDepthBuffer := False;

  m_pFont := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall := CD3DFont.Create('Arial', 9, D3DFONT_BOLD);
  m_pWallMesh                 := nil;
  m_pSphereMesh               := nil;
  m_pConeMesh                 := nil;

  // Set up wall/floor mesh resolution.  Try changing m_n and m_m to see
  // how the lighting is affected.
  m_n := 32;
  m_m := 32;
  m_nTriangles := (m_n-1)*(m_m-1)*2;
  m_bShowHelp := False;
  m_bWireframe := False;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HRESULT;
begin
  Result := m_pFont.InitDeviceObjects(m_pd3dDevice);
  if FAILED(Result) then Exit;

  Result := m_pFontSmall.InitDeviceObjects(m_pd3dDevice);
  if FAILED(Result) then Exit;

  Result:= S_OK;
end;



//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restores scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
var
  mtrl: TD3DMaterial9;
  matIdentity: TD3DXMatrixA16;

  matView: TD3DXMatrixA16;
  vFromPt: TD3DXVector3;
  vLookatPt: TD3DXVector3;
  vUpVec: TD3DXVector3;
  matProj: TD3DXMatrixA16;
  fAspect: Single;
  light: TD3DLight9; // Description of the D3D light
begin
  Result:= BuildWallMesh;
  if FAILED(Result) then Exit;

  // Create sphere and cone meshes to represent the lights
  if FAILED(D3DXCreateSphere(m_pd3dDevice, 0.25, 20, 20, m_pSphereMesh, nil)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  if FAILED(D3DXCreateCylinder(m_pd3dDevice, 0.0, 0.25, 0.5, 20, 20, m_pConeMesh, nil)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  // Set up a material
  D3DUtil_InitMaterial(mtrl, 1.0, 1.0, 1.0);
  m_pd3dDevice.SetMaterial(mtrl);

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse);

  // Set the world matrix
  D3DXMatrixIdentity(matIdentity);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matIdentity);

  // Set the view matrix.
  vFromPt   := D3DXVector3(-10, 10, -10);
  vLookatPt := D3DXVector3(0.0, 0.0, 0.0);
  vUpVec    := D3DXVector3(0.0, 1.0, 0.0);
  D3DXMatrixLookAtLH(matView, vFromPt, vLookatPt, vUpVec);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

  // Set the projection matrix
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Turn on lighting.
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);

  // Enable ambient lighting to a dim, grey light, so objects that
  // are not lit by the other lights are not completely black
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,
    D3DCOLOR_COLORVALUE(0.25, 0.25, 0.25, 1.0));

  // Set light #0 to be a simple, faint grey directional light so
  // the walls and floor are slightly different shades of grey
  ZeroMemory(@light, SizeOf(light));
  light._Type := D3DLIGHT_DIRECTIONAL;
  light.Direction := D3DXVector3(0.3, -0.5, 0.2);
  light.Diffuse.r := 0.25;
  light.Diffuse.g := 0.25;
  light.Diffuse.b := 0.25;
  m_pd3dDevice.SetLight(0, light);

  // Set light #1 to be a simple, bright directional light to use
  // on the mesh representing light #2
  ZeroMemory(@light, SizeOf(light));
  light._Type := D3DLIGHT_DIRECTIONAL;
  light.Direction := D3DXVector3(0.5, -0.5, 0.5);
  light.Diffuse.r := 1.0;
  light.Diffuse.g := 1.0;
  light.Diffuse.b := 1.0;
  m_pd3dDevice.SetLight(1, light);

  // Light #2 will be the light used to light the floor and walls.  It will
  // be set up in FrameMove() since it changes every frame.

  // Restore the font
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: BuildWallMesh()
// Desc: Builds the wall mesh based on m_m and m_n.
//-----------------------------------------------------------------------------
function CMyD3DApplication.BuildWallMesh: HRESULT;
var
  v: PMyVertexArray;
  pWallMeshTemp: ID3DXMesh;
  pdwAdjacency: PDWORD;
  dX, dZ: Single;
  k, z, x: LongWord;
  pIndex: PWord;
  iIndex: Word;
label
  _End;
begin
  v := nil;
  pWallMeshTemp := nil;
  pdwAdjacency := nil;

  SAFE_RELEASE(m_pWallMesh);

  // Create a square grid m_n*m_m for rendering the wall
  Result := D3DXCreateMeshFVF(m_nTriangles, m_nTriangles * 3,
               D3DXMESH_SYSTEMMEM, MYVERTEX_FVF, m_pd3dDevice, pWallMeshTemp);
  if FAILED(Result) then goto _End;

  // Fill in the grid vertex data
  Result:= pWallMeshTemp.LockVertexBuffer(0, Pointer(v));
  if FAILED(Result) then goto _End;

  dX := 1.0/(m_n-1);
  dZ := 1.0/(m_m-1);
  k := 0;
  for z := 0 to (m_m-2) do
  begin
    for x := 0 to (m_n-2) do
    begin
      v[k].p := D3DXVector3(10 * x*dX, 0.0, 10 * z*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * x*dX, 0.0, 10 * (z+1)*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * (z+1)*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * x*dX, 0.0, 10 * z*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * (z+1)*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * z*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
    end;
  end;
  pWallMeshTemp.UnlockVertexBuffer;

  // Fill in index data
  Result:= pWallMeshTemp.LockIndexBuffer(0, Pointer(pIndex));
  if FAILED(Result) then goto _End;

  for iIndex := 0 to m_nTriangles * 3 - 1 do
  begin
    pIndex^:= iIndex;
    Inc(pIndex);
  end;
  pWallMeshTemp.UnlockIndexBuffer;

  // Eliminate redundant vertices
  try
    GetMem(pdwAdjacency, SizeOf(DWORD) * 3 * m_nTriangles);
  except
    Result := E_OUTOFMEMORY;
  end;
  if FAILED(Result) then goto _End;

  Result := D3DXWeldVertices(pWallMeshTemp, D3DXWELDEPSILONS_WELDALL, nil, nil, pdwAdjacency, nil, nil);
  if FAILED(Result) then goto _End;

  // Optimize the mesh
  Result := pWallMeshTemp.OptimizeInplace(D3DXMESHOPT_COMPACT or D3DXMESHOPT_VERTEXCACHE,
             pdwAdjacency, nil, nil, nil);
  if FAILED(Result) then goto _End;

  // Copy the mesh into fast write-only memory
  Result := pWallMeshTemp.CloneMeshFVF(D3DXMESH_WRITEONLY, MYVERTEX_FVF,
              m_pd3dDevice, m_pWallMesh);
  if FAILED(Result) then goto _End;

_End:
  SAFE_RELEASE(pWallMeshTemp);
  FreeMem(pdwAdjacency);
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
var
  dwCaps: DWORD;
  x, y, z: Single;
begin
  ZeroMemory(@m_light, SizeOf(m_light));

  // Rotate through the various light types
  m_light._Type := TD3DLightType(1+((Trunc(m_fTime) div 5) mod 3));

  // Make sure the light type is supported by the device.  If
  // D3DVTXPCAPS_POSITIONALLIGHTS is not set, the device does not support
  // point or spot lights, so change light #2's type to a directional light.
  dwCaps := m_d3dCaps.VertexProcessingCaps;
  if (0 = dwCaps and D3DVTXPCAPS_POSITIONALLIGHTS) then
  begin
    if (m_light._Type = D3DLIGHT_POINT) or (m_light._Type = D3DLIGHT_SPOT)
    then m_light._Type := D3DLIGHT_DIRECTIONAL;
  end;

  // Values for the light position, direction, and color
  x := Sin(m_fTime*2.000);
  y := Sin(m_fTime*2.246);
  z := Sin(m_fTime*2.640);

  m_light.Diffuse.r  := 0.5 + 0.5 * x;
  m_light.Diffuse.g  := 0.5 + 0.5 * y;
  m_light.Diffuse.b  := 0.5 + 0.5 * z;
  m_light.Range      := 100.0;

  case m_light._Type of
    D3DLIGHT_POINT:
    begin
      D3DXVec3Scale(m_light.Position, D3DXVector3(x, y, z),  4.5);
      m_light.Attenuation1 := 0.4;
    end;

    D3DLIGHT_DIRECTIONAL:
      m_light.Direction    := D3DXVector3(x, y, z);

    D3DLIGHT_SPOT:
    begin
      D3DXVec3Scale(m_light.Position, D3DXVector3(x, y, z),  2.0);
      m_light.Direction    := D3DXVector3(x, y, z);
      m_light.Theta        := 0.5;
      m_light.Phi          := 1.0;
      m_light.Falloff      := 1.0;
      m_light.Attenuation0 := 1.0;
    end;
  end;
  m_pd3dDevice.SetLight(2, m_light);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HRESULT;
var
  matWorld, matTrans, matRotate: TD3DXMatrixA16;
  vecFrom, vecAt: TD3DXVector3;
  vecUp: TD3DXVector3;
  matWorldInv: TD3DXMatrixA16;
  strLight: PChar;
  strInfo: array[0..299] of Char;
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET, $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Turn on light #0 and #2, and turn off light #1
    m_pd3dDevice.LightEnable(0, True);
    m_pd3dDevice.LightEnable(1, False);
    m_pd3dDevice.LightEnable(2, True);

    if (m_bWireframe) then
      m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME)
    else
      m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

    // Draw the floor
    D3DXMatrixTranslation(matTrans, -5.0, -5.0, -5.0);
    D3DXMatrixRotationZ(matRotate, 0.0);
    D3DXMatrixMultiply(matWorld, matRotate, matTrans);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
    m_pWallMesh.DrawSubset(0);

    // Draw the back wall
    D3DXMatrixTranslation(matTrans, 5.0, -5.0, -5.0);
    D3DXMatrixRotationZ(matRotate, D3DX_PI/2);
    D3DXMatrixMultiply(matWorld, matRotate, matTrans);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
    m_pWallMesh.DrawSubset(0);

    // Draw the side wall
    D3DXMatrixTranslation(matTrans, -5.0, -5.0, 5.0);
    D3DXMatrixRotationX(matRotate,  -D3DX_PI/2);
    D3DXMatrixMultiply(matWorld, matRotate, matTrans);
    m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
    m_pWallMesh.DrawSubset(0);

    // Turn on light #1, and turn off light #0 and #2
    m_pd3dDevice.LightEnable(0, False);
    m_pd3dDevice.LightEnable(1, True);
    m_pd3dDevice.LightEnable(2, False);

    m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

    // Draw the mesh representing the light
    if (m_light._Type = D3DLIGHT_POINT) then
    begin
      // Just position the point light -- no need to orient it
      with m_light.Position do D3DXMatrixTranslation(matWorld, x, y, z);
      m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
      m_pSphereMesh.DrawSubset(0);
    end else
    begin
      // Position the light and point it in the light's direction
      vecFrom:= D3DXVector3(m_light.Position.x, m_light.Position.y, m_light.Position.z);
      vecAt:= D3DXVector3(m_light.Position.x + m_light.Direction.x,
                          m_light.Position.y + m_light.Direction.y,
                          m_light.Position.z + m_light.Direction.z);
      vecUp:= D3DXVector3(0, 1, 0);
      D3DXMatrixLookAtLH(matWorldInv, vecFrom, vecAt, vecUp);
      D3DXMatrixInverse(matWorld, nil, matWorldInv);
      m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
      m_pConeMesh.DrawSubset(0);
    end;

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);
    case m_light._Type of
      D3DLIGHT_POINT: strLight := 'Point Light';
      D3DLIGHT_SPOT : strLight := 'Spot Light';
     else
       strLight := 'Directional Light';
    end;
    StrFmt(strInfo, '%s, %dx%d vertices in wall/floor mesh', [strLight, m_n, m_m]);
    m_pFontSmall.DrawText( 2, 40, D3DCOLOR_ARGB(255,255,255,255), strInfo);
    if m_bShowHelp then 
    begin
      m_pFontSmall.DrawText( 2, 60, D3DCOLOR_ARGB(255,255,255,255),
        'Use arrow keys to change mesh density');
      m_pFontSmall.DrawText( 2, 80, D3DCOLOR_ARGB(255,255,255,255),
        'Press W to toggle wireframe mode');
    end else
    begin
      m_pFontSmall.DrawText( 2, 60, D3DCOLOR_ARGB(255,255,255,255),
        'Press F1 for help');
    end;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Invalidates device objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
begin
  SAFE_RELEASE(m_pWallMesh);
  SAFE_RELEASE(m_pSphereMesh);
  SAFE_RELEASE(m_pConeMesh);
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;

  Result:= S_OK;
end;




//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;

  Result:= S_OK;
end;




//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
begin
  // Cleanup D3D font
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pFontSmall);

  Result:= S_OK;
end;



//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: LongWord;
  wParam: WPARAM; lParam: LPARAM): LRESULT;
begin
  case uMsg of
    WM_COMMAND:
    case LOWORD(wParam) of
      IDM_TOGGLEHELP:       m_bShowHelp := not m_bShowHelp;
      IDM_TOGGLEWIREFRAME:  m_bWireframe := not m_bWireframe;
      IDM_INCREASERESOLUTION:
        if (m_n < 64) and (m_m < 64) then
        begin
          m_n := m_n *2;
          m_m := m_m *2;
          m_nTriangles := (m_n-1) * (m_m-1) * 2;
          BuildWallMesh;
        end;
      IDM_DECREASERESOLUTION:
        if (m_n > 2) and (m_m > 2) then
        begin
          m_n := m_n div 2;
          m_m := m_m div 2;
          m_nTriangles := (m_n-1) * (m_m-1) * 2;
          BuildWallMesh;
        end;
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

